// $Id: IndividualNearestNeighbours.h 1125 2020-07-10 17:53:09Z ge $
/// \file IndividualNearestNeighbours.h
/// \brief contains the Gibbs Free Energy function class
///
/// $Revision: 1125 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_INDIVIDUALNEARESTNEIGHBOURS_H
#define GBC_INDIVIDUALNEARESTNEIGHBOURS_H "$Id: IndividualNearestNeighbours.h 1125 2020-07-10 17:53:09Z ge $"
#include "NeighbourSequence.h"
#include "Duplex.h"
#include <valarray>
#include <fstream>
#include "ParameterMap.h"
#include <boost/regex.hpp>

using namespace gbc;
namespace gbc
  {
  template<class _Tp=double>
  /// \brief Calculate Gibbs Free Energies and Melting Temperatures
  ///
  /// This implements DNA melting temperature calculations based on the independent nearest neighbour model (INN)
  /// See santalucia96, santalucia98, gray97, gray97b
  class IndividualNearestNeighbours
    {
    public:
      typedef _Tp                              value_type;
      typedef ParameterMap<value_type>         parameter_map_type;
      typedef std::set<std::string>            nearest_neighbours_map_type;
      typedef std::map<std::string,value_type> enthalpy_map_type; //Despite its name this typedef is also used for entropy
      typedef std::valarray<value_type> vector_type;
      typedef SequenceInfo<false,value_type>   sequenceinfo_type;//false refers to periodicity
      typedef typename sequenceinfo_type::base_pair_type     
                                               base_pair_type;
      typedef typename sequenceinfo_type::neighboursequence_type
                                               neighboursequence_type;
      typedef typename sequenceinfo_type::neighbours_ref_type
                                               neighbours_ref_type;
      typedef std::deque<std::string>          string_deque_type;
      typedef typename sequenceinfo_type::salt_map_type
                                               salt_map_type;

      constexpr static const value_type R=1.987;   ///< Gas constant, cal/(K mol)

      sequenceinfo_type*  Sequence_info;
      bool                Debug;
      value_type          Sequence_enthalpy;      ///< Contains the calculated total sequence ethalphy
      value_type          Sequence_entropy;       ///< Contains the calculated total sequence entropy
      value_type          Sequence_terminal_enthalpy;      ///< Contains the calculated terminal sequence ethalphy
      value_type          Sequence_terminal_entropy;       ///< Contains the calculated total sequence entropy
      value_type          Initiation_CG_entropy;  ///< For sequences that contain at least one CG base-pair
      value_type          Initiation_CG_enthalpy; ///< For sequences that contain at least one CG base-pair
      value_type          Initiation_AT_entropy;  ///< For sequences that are AT only
      value_type          Initiation_AT_enthalpy; ///< For sequences that are AT only
      value_type          Initiation_entropy;     ///< Entropy for initiation
      value_type          Initiation_enthalpy;    ///< Enthalpy for initiation
      value_type          Symmetry_correction_entropy;    ///< Symmetry correction of self-complementary sequences
      value_type          Symmetry_correction_enthalpy;    ///< Symmetry correction of self-complementary sequences
      value_type          Terminal_5TA3_enthalpy; ///< AT 5' terminal correction
      value_type          Terminal_AU_enthalpy;   ///< Contribution for AU or UA at ends
      value_type          Terminal_AU_entropy;    ///< Contribution for AU or UA at ends
      value_type          Terminal_AT_enthalpy;   ///< Contribution for AT or TA at ends
      value_type          Terminal_AT_entropy;    ///< Contribution for AT or TA at ends
      value_type          Terminal_CG_enthalpy;   ///< Contribution for AT or TA at ends
      value_type          Terminal_CG_entropy;    ///< Contribution for AT or TA at ends
      value_type          Terminal_dArU_enthalpy;
      value_type          Terminal_dArU_entropy;
      value_type          Terminal_dTrA_enthalpy;
      value_type          Terminal_dTrA_entropy;
      value_type          Terminal_dCrG_enthalpy;
      value_type          Terminal_dCrG_entropy;
      value_type          Terminal_dGrC_enthalpy;
      value_type          Terminal_dGrC_entropy;
      salt_map_type       Parameter_salt_concentration; ///< the salt concentration for which the parameters are valid
      parameter_map_type  parameter_map;          ///< Contains all parameters
      enthalpy_map_type   Enthalpies, Entropies;  ///< Contains all nearest-neighbour enthalpies and entropies
      bool                NN_parameters_only;

    IndividualNearestNeighbours(void) 
        : Debug(false), 
          Sequence_enthalpy(), Sequence_entropy(), Sequence_terminal_enthalpy(), Sequence_terminal_entropy(),
          Initiation_CG_entropy(), Initiation_CG_enthalpy(), Initiation_AT_entropy(), Initiation_AT_enthalpy(),
          Symmetry_correction_entropy(), Symmetry_correction_enthalpy(),
          Terminal_5TA3_enthalpy(),
          Terminal_AU_enthalpy(), Terminal_AU_entropy(), Terminal_CG_enthalpy(), Terminal_CG_entropy(),
          Terminal_dArU_enthalpy(), Terminal_dArU_entropy(), Terminal_dTrA_enthalpy(), Terminal_dTrA_entropy(),
          Terminal_dCrG_enthalpy(), Terminal_dCrG_entropy(), Terminal_dGrC_enthalpy(), Terminal_dGrC_entropy(),
          NN_parameters_only(false) {}
          
//     ~IndividualNearestNeighbours(void)
//       {
//       if (Sequence_info)
//         {
//         int n;
//         for(n=0; n <  Sequence_info->Reduced_neighbours_set_size; ++n)
//           {
//           Sequence_info->ppEnthalpy[n]=Sequence_info->ppEntropy[n]=NULL;
//           }
//         }
//       }

    inline void reset_calculation(void)
      {
      Sequence_enthalpy=value_type();
      Sequence_entropy=value_type();
      }

    /// \brief dummy function, needed in ParameterSearch, has no function
    inline void temperature(value_type T) {}

    /// \brief dummy function, needed in ParameterSearch, has no function
    inline value_type temperature(void) {return value_type();}

    /// \brief dummy function, needed in ParameterSearch, has no function
    friend std::ostream& operator<<(std::ostream &out, const IndividualNearestNeighbours& tm)
      {
      out << tm.parameter_map;
      return out;
      }

    inline bool get_parameters_from_file(const std::string Filename)
      {
      parameter_map.read_from(Filename);
       set_TM_parameters();
      return true;
      }
      
    /// \brief Reads a deque of parameter files and assigns to all maps
    /// All parameters are places into parameter_map.
    inline bool get_parameters_from_list_of_files(string_deque_type &par_file_deque)
      {
      parameter_map.read_from_list_of_files(par_file_deque);
      set_TM_parameters();
      return true;
      }


    void  set_TM_parameters(void) 
      {
      nearest_neighbours_map_type nn;
      parameter_map.get_all_id(nn); // Here we get all combinations of NN from the parameter file
      typename nearest_neighbours_map_type::iterator it;
      boost::regex expr{".+_.+"};
      for(it=nn.begin(); it != nn.end(); ++it)
        {
        if (boost::regex_match(*it, expr))
          {
          std::string parameter_name=*it + ":" + "enthalpy";
          Enthalpies[*it]=parameter_map.get_parameter_or_generic(parameter_name);
          
          parameter_name=*it + ":" + "entropy";
          Entropies[*it] =parameter_map.get_parameter_or_generic(parameter_name);
          }
        }
      Initiation_entropy=parameter_map.get_if_exists("initiation_entropy");
        
      Initiation_enthalpy=parameter_map.get_if_exists("initiation_enthalpy");

      Initiation_CG_entropy=parameter_map.get_if_exists("initiation_CG_entropy");
      
      Initiation_CG_enthalpy=parameter_map.get_if_exists("initiation_CG_enthalpy");

      Initiation_AT_entropy=parameter_map.get_if_exists("initiation_AT_entropy");
	
      Initiation_AT_enthalpy=parameter_map.get_if_exists("initiation_AT_enthalpy");

      Symmetry_correction_entropy=parameter_map.get_if_exists("symmetry_correction_entropy");
      
      Symmetry_correction_enthalpy=parameter_map.get_if_exists("symmetry_correction_enthalpy");
      
      Terminal_5TA3_enthalpy=parameter_map.get_if_exists("terminal_5TA3_enthalpy");
      
      Terminal_AU_enthalpy=parameter_map.get_if_exists("terminal_AU_enthalpy");

      Terminal_AU_entropy=parameter_map.get_if_exists("terminal_AU_entropy");

      Terminal_AT_enthalpy=parameter_map.get_if_exists("terminal_AT_enthalpy");

      Terminal_AT_entropy=parameter_map.get_if_exists("terminal_AT_entropy");

      Terminal_CG_enthalpy=parameter_map.get_if_exists("terminal_CG_enthalpy");

      Terminal_CG_entropy=parameter_map.get_if_exists("terminal_CG_entropy");

      Terminal_dArU_enthalpy=parameter_map.get_if_exists("terminal_dArU_enthalpy");
      Terminal_dArU_entropy=parameter_map.get_if_exists("terminal_dArU_entropy");
      
      Terminal_dTrA_enthalpy=parameter_map.get_if_exists("terminal_dTrA_enthalpy");
      Terminal_dTrA_entropy=parameter_map.get_if_exists("terminal_dTrA_entropy");
      
      Terminal_dCrG_enthalpy=parameter_map.get_if_exists("terminal_dCrG_enthalpy");
      Terminal_dCrG_entropy=parameter_map.get_if_exists("terminal_dCrG_entropy");

      Terminal_dGrC_enthalpy=parameter_map.get_if_exists("terminal_dGrC_enthalpy");
      Terminal_dGrC_entropy=parameter_map.get_if_exists("terminal_dGrC_entropy");

      if ((Enthalpies.size() == size_t()) || (Entropies.size() == size_t())) 
        std::cout << __FILE__ << ":" << __LINE__ << " no specific Enthalpy or Entropy values, falling back to generic values" << std::endl;
      
      Parameter_salt_concentration["Na+"]=parameter_map.get_if_exists("Na+:concentration");
      if (Parameter_salt_concentration["Na+"] != value_type())
        std::cout << __FILE__ << ":" << __LINE__ << " setting parameter salt concentration to " << Parameter_salt_concentration["Na+"] << std::endl;
      }
      
      
    inline value_type get_terminal_parameter (base_pair_type &bp, std::string parameter_name)
      {
      std::string formatted_parameter_name = std::string(bp) + parameter_name;
      
      value_type parameter = parameter_map.get_if_exists(formatted_parameter_name);

      if (parameter != value_type()) 
        {
        if (Debug) 
          std::cerr << "[Debug] " << formatted_parameter_name  << " = " << parameter << std::endl;
        }
      else
        {
        if (Debug) 
          std::cerr << "[Debug] No parameter " << formatted_parameter_name  << " found " << std::endl;
        }
        
      return parameter;
      }
    

    /// \returns Enthalpy in Kcal/mol
    /// \returns Entropy in cal/(mol.K)
    inline void calculate_total_enthalpy_and_entropy(void)
      {
      value_type total_enthalpy=value_type();
      value_type total_entropy =value_type();
      Sequence_terminal_enthalpy = value_type();
      Sequence_terminal_entropy  = value_type();
      
      if (Debug) std::cerr << "Sequence " << Sequence_info->Exact_duplex << std::endl;
      int n;
      if (!Sequence_info->ppEnthalpy[0] or !Sequence_info->ppEntropy[0])
        {
        typename neighbours_ref_type::const_iterator it;
        for(n=0, it=Sequence_info->Reduced_neighbours_set.begin();
            it != Sequence_info->Reduced_neighbours_set.end(); ++it, ++n)
          {
          typename enthalpy_map_type::iterator enth=Enthalpies.find(it->first);
          if (enth == Enthalpies.end())
            {
            std::cerr << __FILE__ << " " << __LINE__ << "\n" << "Enthalpy for " << it->first << " not found" << std::endl;
            std::cerr << "Enthalpies[] has " << Enthalpies.size() << " elements" << std::endl;
            }
          else 
            {
            Sequence_info->ppEnthalpy[n] = &(enth->second);
            total_enthalpy += it->second*enth->second;
            }
          typename enthalpy_map_type::iterator entr=Entropies.find(it->first);
          if (entr == Entropies.end())
            {
            std::cerr << __FILE__ << " " << __LINE__ << "\n" << "Entropy for " << it->first << " not found" << std::endl;
            std::cerr << "Entropies[] has " << Entropies.size() << " elements" << std::endl;
            }
          else 
            {
            Sequence_info->ppEntropy[n] = &(entr->second);
            total_entropy += it->second*entr->second;
            }
          if (Debug) 
            {
            std::cerr << it->first;
            std::cerr << " DH(" << it->first << ")= "<<  Enthalpies[it->first];
            std::cerr << " DS =" <<  Entropies[it->first] << std::endl;
            }
          }
        }
      else
        {
        for(n=0; n < Sequence_info->Reduced_neighbours_set.size(); n++)
          {
          total_enthalpy += Sequence_info->pReduced_nn_number[n] * *(Sequence_info->ppEnthalpy[n]);
          total_entropy  += Sequence_info->pReduced_nn_number[n] * *(Sequence_info->ppEntropy[n] );
          }
        }
      
        
      if (NN_parameters_only) //skip all further initialization contributions
        {
        Sequence_enthalpy=Sequence_info->enthalpy.predicted=total_enthalpy;
        Sequence_entropy=Sequence_info->entropy.predicted=total_entropy;
        if (Debug)  std::cerr << "Total DH = " << Sequence_enthalpy << " Total DS = " << Sequence_entropy << std::endl;
        return;
        }
        
      if (Sequence_info->Has_CG)    
        {
        total_enthalpy += Initiation_CG_enthalpy; 
        total_entropy  += Initiation_CG_entropy; 
        if (Debug)
          {
          if (Initiation_CG_enthalpy != 0) std::cerr << " Initiation CG enthalpy= " << Initiation_CG_enthalpy << std::endl;
          if (Initiation_CG_entropy != 0)  std::cerr << " Initiation CG entropy= " << Initiation_CG_entropy << std::endl;
          }
        }  
      if (Sequence_info->Is_AT_only)   
        {
        total_enthalpy += Initiation_AT_enthalpy;
        total_entropy += Initiation_AT_entropy;
        if (Debug)
          {
          if (Initiation_AT_enthalpy != 0) std::cerr << " Initiation AT enthalpy = " << Initiation_AT_enthalpy << std::endl;
          if (Initiation_AT_entropy != 0)  std::cerr << " Initiation AT entropy = " << Initiation_AT_entropy << std::endl;
          }
        }
        
      Sequence_terminal_enthalpy += get_terminal_parameter(Sequence_info->Terminal3_bp_reduced,std::string(":terminal_enthalpy"));
      Sequence_terminal_entropy  += get_terminal_parameter(Sequence_info->Terminal3_bp_reduced,std::string(":terminal_entropy"));

      Sequence_terminal_enthalpy += get_terminal_parameter(Sequence_info->Terminal5_bp_reduced,std::string(":terminal_enthalpy"));
      Sequence_terminal_entropy  += get_terminal_parameter(Sequence_info->Terminal5_bp_reduced,std::string(":terminal_entropy"));
      
      base_pair_type bp=Sequence_info->Terminal3_bp;
      bp.swap();
              
      Sequence_terminal_enthalpy += get_terminal_parameter(bp,std::string(":53_terminal_enthalpy"));
      Sequence_terminal_entropy  += get_terminal_parameter(bp,std::string(":53_terminal_entropy"));

      Sequence_terminal_enthalpy += get_terminal_parameter(Sequence_info->Terminal5_bp,std::string(":53_terminal_enthalpy"));
      Sequence_terminal_entropy  += get_terminal_parameter(Sequence_info->Terminal5_bp,std::string(":53_terminal_entropy"));
      

      if (Sequence_info->Terminal_5TA3)
        {
        Sequence_terminal_enthalpy += Sequence_info->Terminal_5TA3*Terminal_5TA3_enthalpy;
        if (Debug && (Terminal_5TA3_enthalpy != 0)) 
          std::cerr << "Terminal_5TA3_enthalpy = " << Sequence_info->Terminal_5TA3 << " X " << Terminal_5TA3_enthalpy << " = " << Sequence_info->Terminal_5TA3*Terminal_5TA3_enthalpy << std::endl;
        }
      if (Sequence_info->Terminal_AU)
        {
        Sequence_terminal_enthalpy += Sequence_info->Terminal_AU*Terminal_AU_enthalpy;
        Sequence_terminal_entropy  += Sequence_info->Terminal_AU*Terminal_AU_entropy;
        if (Debug)
          {
          if (Terminal_AU_enthalpy != 0) 
          std::cerr << "Terminal_AU_enthalpy = " << Sequence_info->Terminal_AU << " X " << Terminal_AU_enthalpy << " = " << Sequence_info->Terminal_AU*Terminal_AU_enthalpy << std::endl;
          if (Terminal_AU_entropy != 0)
          std::cerr << "Terminal_AU_entropy = " << Sequence_info->Terminal_AU << " X " << Terminal_AU_entropy << " = "  << Sequence_info->Terminal_AU*Terminal_AU_entropy << std::endl;
          }
        }
       if (Sequence_info->Terminal_AT)
        {
        Sequence_terminal_enthalpy += Sequence_info->Terminal_AT*Terminal_AT_enthalpy;
        Sequence_terminal_entropy  += Sequence_info->Terminal_AT*Terminal_AT_entropy;
        if (Debug)
          {
          if (Terminal_AT_enthalpy != 0)
          std::cerr << "Terminal_AT_enthalpy = " << Sequence_info->Terminal_AT << " X " << Terminal_AT_enthalpy << " = " << Sequence_info->Terminal_AT*Terminal_AT_enthalpy << std::endl;
          if (Terminal_AT_entropy != 0)
          std::cerr << "Terminal_AT_entropy = " << Sequence_info->Terminal_AT << " X " << Terminal_AT_entropy << " = "  << Sequence_info->Terminal_AT*Terminal_AT_entropy << std::endl;
          }
        }
      if (Sequence_info->Terminal_CG)
        {
        Sequence_terminal_enthalpy += Sequence_info->Terminal_CG*Terminal_CG_enthalpy;
        Sequence_terminal_entropy  += Sequence_info->Terminal_CG*Terminal_CG_entropy;
        if (Debug)
          {
          if (Terminal_CG_enthalpy != 0)
          std::cerr << "Terminal_CG_enthalpy = " << Sequence_info->Terminal_CG << " X " << Terminal_CG_enthalpy << " = " << Sequence_info->Terminal_CG*Terminal_CG_enthalpy << std::endl;
          if (Terminal_CG_entropy != 0)
          std::cerr << "Terminal_CG_entropy = " << Sequence_info->Terminal_CG << " X " << Terminal_CG_entropy << " = "  << Sequence_info->Terminal_CG*Terminal_CG_entropy << std::endl;
          }
        }
      if (Sequence_info->Terminal_dArU)
        {
        Sequence_terminal_enthalpy += Sequence_info->Terminal_dArU*Terminal_dArU_enthalpy;
        Sequence_terminal_entropy  += Sequence_info->Terminal_dArU*Terminal_dArU_entropy;
        if (Debug)
          {
          if (Terminal_dArU_enthalpy != 0)
          std::cerr << "Terminal_dArU_enthalpy = " << Sequence_info->Terminal_dArU << " X " << Terminal_dArU_enthalpy << " = " << Sequence_info->Terminal_dArU*Terminal_dArU_enthalpy << std::endl;
          if (Terminal_dArU_entropy != 0)
          std::cerr << "Terminal_dArU_entropy = " << Sequence_info->Terminal_dArU << " X " << Terminal_dArU_entropy << " = "  << Sequence_info->Terminal_dArU*Terminal_dArU_entropy << std::endl;
          }
        }
      if (Sequence_info->Terminal_dTrA)
        {
        Sequence_terminal_enthalpy += Sequence_info->Terminal_dTrA*Terminal_dTrA_enthalpy;
        Sequence_terminal_entropy  += Sequence_info->Terminal_dTrA*Terminal_dTrA_entropy;
        if (Debug)
          {
          if (Terminal_dTrA_enthalpy != 0)
          std::cerr << "Terminal_dTrA_enthalpy = " << Sequence_info->Terminal_dTrA << " X " << Terminal_dTrA_enthalpy << " = " << Sequence_info->Terminal_dTrA*Terminal_dTrA_enthalpy << std::endl;
          if (Terminal_dTrA_entropy != 0)
          std::cerr << "Terminal_dTrA_entropy = " << Sequence_info->Terminal_dTrA << " X " << Terminal_dTrA_entropy << " = "  << Sequence_info->Terminal_dTrA*Terminal_dTrA_entropy << std::endl;
          }
        }


      if (Sequence_info->Terminal_dCrG)
        {
        Sequence_terminal_enthalpy += Sequence_info->Terminal_dCrG*Terminal_dCrG_enthalpy;
        Sequence_terminal_entropy  += Sequence_info->Terminal_dCrG*Terminal_dCrG_entropy;
        if (Debug)
          {
          if (Terminal_dCrG_enthalpy != 0) 
          std::cerr << "Terminal_dCrG_enthalpy = " << Sequence_info->Terminal_dCrG << " X " << Terminal_dCrG_enthalpy << " = " << Sequence_info->Terminal_dCrG*Terminal_dCrG_enthalpy << std::endl;
          if (Terminal_dCrG_entropy != 0)
          std::cerr << "Terminal_dCrG_entropy = " << Sequence_info->Terminal_dCrG << " X " << Terminal_dCrG_entropy << " = "  << Sequence_info->Terminal_dCrG*Terminal_dCrG_entropy << std::endl;
          }
        }
      if (Sequence_info->Terminal_dGrC)
        {
        Sequence_terminal_enthalpy += Sequence_info->Terminal_dGrC*Terminal_dGrC_enthalpy;
        Sequence_terminal_entropy  += Sequence_info->Terminal_dGrC*Terminal_dGrC_entropy;
        if (Debug)
          {
          if (Terminal_dGrC_enthalpy != 0)
          std::cerr << "Terminal_dGrC_enthalpy = " << Sequence_info->Terminal_dGrC << " X " << Terminal_dGrC_enthalpy << " = " << Sequence_info->Terminal_dGrC*Terminal_dGrC_enthalpy << std::endl;
          if (Terminal_dGrC_entropy != 0)
          std::cerr << "Terminal_dGrC_entropy = " << Sequence_info->Terminal_dGrC << " X " << Terminal_dGrC_entropy << " = "  << Sequence_info->Terminal_dGrC*Terminal_dGrC_entropy << std::endl;
          }
        }

      total_enthalpy += Sequence_terminal_enthalpy;
      total_entropy  += Sequence_terminal_entropy;

      if (Initiation_enthalpy != 0)
        {
        total_enthalpy += Initiation_enthalpy;
        if (Debug)  std::cerr << "Initiation_enthalpy = " << Initiation_enthalpy << std::endl;
        }
      if (Initiation_entropy != 0)
        {
        total_entropy += Initiation_entropy;
        if (Debug)  std::cerr << "Initiation_entropy = " << Initiation_entropy << std::endl;
        }
        
      if (Sequence_info->Self_complementary) 
        {
        total_entropy += Symmetry_correction_entropy;
        total_enthalpy += Symmetry_correction_enthalpy;
        if (Debug) 
          {
          std::cerr << "Symmetry correction entropy = " << Symmetry_correction_entropy << std::endl;
          std::cerr << "Symmetry correction enthalpy = " << Symmetry_correction_enthalpy << std::endl;
          }
        }
      
      Sequence_enthalpy=Sequence_info->enthalpy.predicted=total_enthalpy;
      Sequence_entropy=Sequence_info->entropy.predicted=total_entropy;
      if (Debug)  std::cerr << "Total DH = " << Sequence_enthalpy << " Total DS = " << Sequence_entropy << std::endl;
      }

    /// \returns Enthalpy in Kcal/mol
    inline value_type total_enthalpy(void)
      {
      if (Sequence_enthalpy == value_type()) calculate_total_enthalpy_and_entropy();
      return Sequence_enthalpy;
      }

    /// \returns Entropy in cal/(mol.K)
    inline value_type total_entropy(void)
      {
      if (Sequence_entropy == value_type()) calculate_total_enthalpy_and_entropy();
      return Sequence_entropy;
      }

    inline value_type melting_temperature(void)
      {
      value_type alpha=4.0;
      if (Sequence_info->Self_complementary) alpha=1.0;
      if (Sequence_info->species_concentration == value_type())
        std::cerr << __FILE__ << ":" << __LINE__ << " warning: species concentration is zero (Ct=0), temperature calculation will fail due to ln(0) calculation " << std::endl;
      
      //ATTENTION The following equation is what usually is called the "melting temperature" although strictly speaking it is the reverse processes, for a discussion
      //see Footnote 7 of Schreiber-Gosche, Sherrie, and Robert A. Edwards. "Thermodynamics of oligonucleotide duplex melting." Journal of chemical education 86.5 (2009): 644. (26/06/2020) 
      Sequence_info->temperature.predicted=total_enthalpy()*1000.0/(total_entropy()+R*log(Sequence_info->species_concentration*sequenceinfo_type::Concentration_factor/alpha))-273.15;

      if (Sequence_info->Prediction_with_salt_correction) Sequence_info->apply_salt_correction(Parameter_salt_concentration["Na+"]); //If Prediction_with_salt_correction==true and scheme != nocorrection
      return Sequence_info->temperature.predicted;
      }

    ///< \brief Calculates the Gibbs Free Energy (in kcal/mol)
    ///
    /// If the temperature is not given we assume 37 C.
    /// If entropy and ethalpy are not given, we calculate them.
    inline value_type gibbs_free_energy(value_type temp=37.0,
                                        value_type enth=value_type(),
                                        value_type entr=value_type())
      {
      if (enth == value_type()) enth=total_enthalpy();
      if (entr == value_type()) entr=total_entropy();
      return (enth*1000.0-(273.15+temp)*entr)/1000.0;
      }

    };

  };

#endif
